#include <linuxmt/types.h>
#include <linuxmt/config.h>
#include <linuxmt/sched.h>
#include <linuxmt/errno.h>
#include <arch/system.h>

#ifdef CONFIG_BIOS_HD

struct hd_info
{
	unsigned short cyl;
	unsigned short wpcom;
	unsigned short lzone;
	unsigned char head;
	unsigned char sect;
	unsigned char ctl;
};

static struct hd_info disks[2];

/*
 *	Just don't re-enter 8). Easy ways to get the parameters to
 *	the #asm bits.. should use 2[bp] etc - FIXME
 */
 
static unsigned int hd_unit;  /* BIOS xt drive (drive + 0x80) */
static unsigned int hd_stat; /* XT status */
static unsigned int hd_cylh; /* BIOS cylender */
static unsigned int hd_cyll; /* BIOS cylender */
static unsigned int hd_head; /* BIOS head */
static unsigned int hd_sector; /* BIOS sector */

int hd_seek(hd_sec, hd_hed, hd_cyl, hd_dsk)
unsigned int hd_sec;
unsigned int hd_hed;
unsigned int hd_cyl;
unsigned int hd_dsk;
{
	hd_unit=hd_dsk + 0x80;
	hd_head = hd_hed & 0x0f; /* head */
	hd_cylh=((hd_cyl / 256) << 6) | hd_sec;    /* high 2 bits + sector */
	hd_cyll=hd_cyl & 0x00ff; /* low 8 bits */
/* printk("%d %d %d %d\r",hd_hed,hd_cyl,hd_cylh, hd_cyll); */

	/*
	 *	perform seek operation, return status 
	 */
	 
#asm
	mov	ah,#0x0c
	mov	cl,_hd_cylh
	mov     ch,_hd_cyll
	mov	dh,_hd_head
	mov     dl,_hd_unit
	int	0x13
	mov     _hd_stat,ah
#endasm

	/* printk("%d ",hd_stat); */
	return(hd_stat);
}

/*
 *	Recalibrate and recover the drive.
 */
 
int hd_recal(hd_dsk)
unsigned int hd_dsk;
{
	hd_unit = hd_dsk + 0x80;
	/*
	 *	Re-calibrate the drive (home heads) 
	 */
#asm
	mov	ah,#0x11
	mov	dl,_hd_unit
	int	0x13		
	mov	_hd_stat,ah
#endasm
	return(hd_stat);
}

/*
 *	Initialise the hard disk controller.
 */
 
void hd_init()
{
	
	int d=0;
	int pos=0x80;	/* Disk base */
	unsigned int current_head;
	unsigned int current_cyl;
	unsigned int current_sector;
	unsigned int hd_status;
	unsigned long size;
	
	for(d=0;d<2;d++)
	{
		/*
		 *	Pull the AT settings
		 */
		 
		disks[d].cyl	=	peekw(0x9000,pos);
		disks[d].head	=	peekb(0x9000,pos+2);
		disks[d].wpcom	=	peekw(0x9000,pos+5);
		disks[d].ctl	=	peekb(0x9000,pos+8);
		disks[d].lzone	=	peekw(0x9000,pos+12);
		disks[d].sect	=	peekb(0x9000,pos+14);
		
		/*
		 *	At least one person has an XT that doesn't answer
		 *	the sectors value.
		 */
	
		if(arch_cpu<2)
		{
#ifdef CONFIG_XT_BIOS_HD
		 	/* 
		 	 * Ever hear of the "binary chop" for searching a 
		 	 * database?? well, we can do the SAME thing here!
			 * But for now, we will seek cylinders, heads and 
			 * sectors until we FAIL, that will indicate the size!
			 *
			 * Here's the problems:
			 * 1: Old style controllers aren't "smart" enough to 
			 *	let the system know how big they are.
			 * 2: In the "old days" everyone had thier own ideas
			 *	of what is a "standard".
			 *
			 * Here's how we solve this Pandora's box.
			 * 1: Set the geometry to the following...
			 * 1024 cyl, 64 sec, 16 heads
			 * 2: We know that this will not be correct, so we 
			 * SEEK in halves untill we fail and find the "real" end
			 * points of the device!
			 */
			disks[d].sect = 0;	/* Assume the worse case! */
			disks[d].cyl = 0;
			disks[d].head = 0;
			printk("Calculating XT hard-disc geometry");	
				/* Now we  ask the hard disk how big it is */
			hd_status=hd_recal(d);
			if (hd_status != 0)
			{
				/* we encountered a recalibration error, no disk! */
			 	disks[d].sect = 0;
			 	disks[d].cyl =  0;
			 	disks[d].head = 0;
				goto getout;
			}
			printk("."); /* indicate recal is good */
			/* we *KNOW* a hard disc is there! */

			/* seek out cylinders till we FAIL */
			for ( current_cyl=0; current_cyl < 1024 ; current_cyl++)
			{
				hd_status=hd_seek(1,0,current_cyl,d);
				if(hd_status)
				{
					/* try an alternate area */
					hd_status=hd_recal(d);
					hd_status=hd_seek(2,0,current_cyl,d);
					if(hd_status)
						goto exit_cyl;
				}
			} 
		exit_cyl:
			hd_status=hd_recal(d);
			disks[d].cyl = current_cyl-1;
			if (current_cyl <= 0)    /* Houston, we have a problem... */
				goto getout;
			printk(".");

			/* Seek out heads till we FAIL */
			for (current_head=0 ; current_head<16 ; current_head++)
			{
				/* perform seek operation, return status */
				hd_status=hd_seek(1,current_head,0,d);
				if(hd_status)
				{
					/* try an alternate area */
					hd_status=hd_recal(d);
					hd_status=hd_seek(2,current_head,0,d);
					if(hd_status)
						goto exit_head;
				}
			} 
		exit_head:
			hd_status=hd_recal(d);
			disks[d].head = current_head; /* odd... we don't subtract 1 */
			if (current_head <= 0)    /* Houston, we have a problem... */
				goto getout;
			printk(".");		
			/* seek out sectors till we FAIL */
			for ( current_sector=1;current_sector<65; current_sector++)
			{
				hd_status=hd_seek(current_sector,0,0,d);
				if(hd_status) /* sector not found error */
				{
					/* try an alternate area */
					hd_status=hd_recal(d);
					hd_status=hd_seek(current_sector,1,0,d);
					if(hd_status)
						goto exit_sec;
				}
			} 
		exit_sec:
			disks[d].sect = current_sector - 1;
			if (current_sector <= 0)    /* Houston, we have a problem... */
				goto getout;
			printk("OK");
			/* common exit */
		getout:
			printk("\n");
			hd_status=hd_recal(d);
#else
			disks[d].cyl=0;
#endif						
		}
		pos+=16;
		if(disks[d].cyl)
		{
			/* Size in 1K blocks */
			size=disks[d].head*disks[d].sect*(disks[d].cyl/2);
/*			printk("hd%c: %d 1K blocks \n",d+'a',size);*/
			/* Convert to Mb */
			size=size / 1024;
			printk("hd%c: %d heads %d cylinders, %d sectors (%ldMb).\n",
				d+'a',disks[d].head,disks[d].cyl,disks[d].sect, size);
		}
		
	}
}	
#endif
